--                              -*- Mode: Ada -*-
-- Filename        : $Source: /home/ACDE/ACDE/ASCL/RCS/ascl-app-initfile.adb,v $
-- Description     : handle initfiles
-- Author          : Michael Erdmann
-- Created On      : Sat May 23 17:37:01 1998
-- Last Modified By: $Author: ACDE $
-- Last Modified On: $Date: 2000/01/23 09:48:34 $
-- Status          : $State: Released $
--
-- Functional Description
-- ======================
--
--
-- History
-- =======
-- $Log: ascl-app-initfile.adb,v $
-- Revision 1.2  2000/01/23 09:48:34  ACDE
-- Release Version base_1_1
--
-- Revision 1.1  2000/01/23 09:46:34  ACDE
-- Release Version 1_0
--
-- Revision 1.12  1999/11/27 16:56:18  ACDE
-- Release Version 1P010
--
-- Revision 1.1  1999/11/27 09:33:25  ACDE
-- Initial revision
--
-- Revision 1.1  1999/08/06 16:58:29  erdmann
-- No comments
--
--
--
with Ada.Text_IO;             use Ada.Text_IO;
with Ada.Exceptions;          use Ada.Exceptions;
with Ada.Characters.Handling; use Ada.Characters.Handling;
with Ada.Strings;             use Ada.Strings;
with Ada.Strings.Fixed;       use Ada.Strings.Fixed;
with Unchecked_Deallocation;

package body ASCL.App.Initfile is

   ---====================================================================---
   ---===         L O C A L     D E C A L R A T I O N S                ===---
   ---====================================================================---
   subtype Identifier is String(1..40);
   subtype Value_String is String(1..130);

   Blank_Identifier : constant Identifier := (others=>' ');

   type Variable;
   type Variable_Access is access Variable;

   type Variable is record
         Next    : Variable_Access := null;
         Name    : Identifier;
         Section : Identifier;
         Value   : Value_String;
      end record;

   --|
   --| I N S T A N C E   D A T A     O F    T H I S    P A C K A G E
   --|
   type Object_Data is record
         First           : Variable_Access := null;
         Next            : Variable_Access := null;
         Current_Section : Identifier;
         File_Name       : String(1..256);
      end record;


   ---====================================================================---
   ---===               L O C A L   M E T H O D S                      ===---
   ---====================================================================---

   procedure Free is
            new Unchecked_Deallocation( Object_Data, Object_Data_Access);

   procedure Free is
            new Unchecked_Deallocation( Variable, Variable_Access);

   --|
   --| Initialize the object data
   --|
   function Initialize return Object_Data_Access is
      Result : Object_Data_Access := new Object_Data;
   begin
      return Result;
   end Initialize;


   --|
   --| read in the defintion file and insert the found data
   --| into the list of definitions.
   --|
   procedure Read_File( O : in Object; File_Name : in String ) is
      This        : Object_Data_Access := O.Data;
      Src         : String(1..512);
      Input       : File_Type;
      Section     : Identifier;
      Name        : Identifier;
      Value       : Value_String;
      Line_Length : Natural;
      Line_Nbr    : Integer := 0;

      procedure Handle_Section is
      begin
         Section := Blank_Identifier;
         for I in 2..Section'Length loop
            exit when Src(I)=']';

            Section(I-1) := Src(I);
         end loop;
      end Handle_Section;

      procedure Handle_Assignement is
         type Syntax_State is ( S_Value, S_Name );

         State : Syntax_State := S_Name;
         Pos   : Positive := 1;
         C     : Positive := 1;
      begin
         Name := Blank_Identifier;
         Value := (others=>' ');

         while Pos <= Line_Length loop
            exit when ( Src(Pos) = ';' or Src(Pos) = '#' );

            case State is
               when S_Name =>
                  if Src(Pos) = '=' then
                     State := S_Value;
                     C := 1;
                  else
                     if Is_Graphic(Src(Pos)) then
                        Name(C) := Src(Pos);
                        C := C + 1;
                     end if;
                  end if;
               when S_Value =>
                  Value(C) := Src(Pos);
                  C := C + 1;
            end case;
            Pos := Pos + 1;
         end loop;
      end Handle_Assignement;

      procedure Insert_Assignement is
         New_Value : Variable_Access := new Variable;
      begin
         New_Value.Next := This.First;
         This.First := New_Value;

         New_Value.Section := Section;
         New_Value.Name    := Name;
         New_Value.Value   := Value;
      end Insert_Assignement;


   begin
      Open( Input, In_File, File_Name );
      if not Is_Open( Input ) then
          Raise_Exception(File_Not_Found'Identity,
                     "Initfile.Read_file: " & File_Name);
      end if;

      while not End_Of_File(Input) loop
         Src := (others=>' ');

         Get_Line( Input, Src, Line_Length );
         Line_Nbr := Line_Nbr + 1;

         case Src(1) is
            when '#'|';' =>             -- comment line
               null;
            when '['     =>             -- section line
               Handle_Section;
            when others  =>             -- any value assignment
               if Is_Letter(Src(1)) then
                  Handle_Assignement;
                  Insert_Assignement;
               elsif Line_Length > 0 then
                  Close(Input);
                  Raise_Exception(
                     Syntax_Error'Identity,
                     "file: " & Name &
                     "at line" & Positive_Count'Image(Line(Input) ) );
               end if;
         end case;
      end loop;
      Close(Input);

   exception
      when The_Error : others =>
         Raise_Exception(Exception_Identity(The_Error),
                     " Initfile.Read_file: " & File_Name &
                     " at line:" & Integer'Image(Line_Nbr) &
                     " length " & Integer'Image(Line_Length));
   end Read_File;
   ---====================================================================---
   ---===            E X P O R T E D    M E T H O D S                  ===---
   ---====================================================================---

   --|
   --| Initialize the object instance data. This implies no
   --| functionality regarding the
   --|
   procedure Initialize( This : in out Object ) is
      O : Object_Data_Access;
   begin
      if This.Data = null then
         This.Data := Initialize;
      end if;
      O := This.Data;

      O.First := null;
   end Initialize;

   --|
   --| Finalize the initfile object, by delting the variable stored
   --| and removing the pointer towards the object data.
   --|
   procedure Finalize( This : in out Object ) is
      O : Object_Data_Access := This.Data;
      V : Variable_Access := O.First;
   begin
      while V /= null loop
         O.First := V.Next;
         free(v);
         V := O.First;
      end loop;

      Free( O );
      This.Data := null;
   end Finalize;

   --|
   --| Get the contents of the init file.
   --|
   procedure Get( This : in Object; Name : in String := "" ) is
      O : Object_Data_Access := This.Data;
   begin
      if O = null then
         Raise_Exception( Usage_Error'Identity,
                          "Initfile.Get: Object data not initialized");
      end if;
      if Name /= "" then
         Read_File( This, Name );
         Move( Name, O.File_Name );
      else
         Read_File( This, O.File_Name );
      end if;
   end Get;


   --|
   --| set the current section in the initial data file
   --|
   procedure Section( This : in Object; Name : String ) is
      O : Object_Data_Access := This.Data;
   begin
      if O = null then
         Raise_Exception( Usage_Error'Identity,
                          "Initfile.Section: Object data not initialized");
      end if;
      Move( Name, O.Current_Section );
   end Section;

   --|
   --| Get the value of a given variable in the current section
   --|
   function Value( This  : in Object; Name : in String ) return String is
      O         : Object_Data_Access := This.Data;
      V         : Variable_Access    := O.First;
      AName     : Identifier         := Blank_Identifier;
      Result    : Value_String       := (others=>' ');

      function match( V: in Variable_Access ) return Boolean is
      begin
         return aName = V.Name and O.Current_Section = V.Section;
      end match;

   begin
      if O = null then
         Raise_Exception(
              Usage_Error'Identity,
              "Initfile.Value: Object data not initialized");
      end if;

      Move( Name, AName );
      while V /= null loop
         if match( V ) then
            Move( V.Value, Result );
            return Trim( Result, Both );
         end if;
         V := V.Next;
      end loop;
      Raise_Exception(Value_Not_found'Identity,
                      "Initfile.Value: " &
                      "Section : " & O.Current_Section &
                      "Name : " & Name );
   end Value;


   --|
   --| get all defintions stored in a single section of the
   --| init files.
   --|
   procedure Next( This   : in Object;
                   Name   : out String;
                   Result : out String;
                   Last   : out Boolean;
                   First  : in Boolean :=False )
   is
      O      : Object_Data_Access := This.Data;
      V      : Variable_Access    := null;
   begin
      Last := False;

      if First then
         O.Next := O.First;
      end if;


      while O.Next /= null loop
         V := O.Next;
         if V.Section = O.Current_Section then
            Move( V.Value, Result );
            Move( V.Name,  Name );
            O.Next := V.Next;
            return;
         end if;

         O.Next := V.Next;
      end loop;
      Last := True;
   end Next;

end ASCL.App.Initfile;
