--|
--| Filename        : $Source: /home/ACDE/ACDE/ASCL/RCS/ascl-debugging_support.adb,v $
--| Description     : Base class for debugging support
--| Author          : Michael Erdmann
--| Created On      : 25.3.1999
--| Last Modified By: $Author: ACDE $
--| Last Modified On: $Date: 2000/01/23 09:48:34 $
--| Status          : $State: Released $
--|
--| Functional Description
--| ======================
--|
--| This module provide the infrstructure of the debugging/tracing
--| within an ada library. The procedure output which has to be
--| supplied by the extension is the gateway for the debuging
--| information. It has to be adopted to the environment which
--| is used.
--|
--| Restrictions
--| ============
--|
--| References
--| ==========
--|
--|
--| History
--| =======
--| $Log: ascl-debugging_support.adb,v $
--| Revision 1.2  2000/01/23 09:48:34  ACDE
--| Release Version base_1_1
--|
--| Revision 1.1  2000/01/23 09:46:34  ACDE
--| Release Version 1_0
--|
--| Revision 1.13  1999/11/27 16:56:18  ACDE
--| Release Version 1P010
--|
--| Revision 1.12  1999/11/27 16:50:10  ACDE
--| Preinstallation checkin
--|
--| Revision 1.11  1999/11/27 16:45:46  ACDE
--| Release Version 1P009
--|
--| Revision 1.10  1999/11/27 16:44:56  ACDE
--| Preinstallation checkin
--|
--| Revision 1.9  1999/11/27 16:42:27  ACDE
--| Preinstallation checkin
--|
--| Revision 1.8  1999/11/27 16:38:33  ACDE
--| Preinstallation checkin
--|
--| Revision 1.7  1999/11/27 16:35:39  ACDE
--| Preinstallation checkin
--|
--| Revision 1.6  1999/11/27 16:33:28  ACDE
--| Preinstallation checkin
--|
--| Revision 1.5  1999/11/27 16:33:00  ACDE
--| Preinstallation checkin
--|
--| Revision 1.4  1999/11/27 16:30:28  ACDE
--| Preinstallation checkin
--|
--| Revision 1.3  1999/11/27 15:47:05  ACDE
--| Installation checkin
--|
--| Revision 1.2  1999/11/27 09:33:55  ACDE
--| sg
--|
--| Revision 1.1  1999/11/26 14:37:14  ACDE
--| error corrections
--|
--| Revision 1.3  1999/08/16 18:20:04  erdmann
--| Test complete
--|

--* Ada
with Ada.Exceptions;               use Ada.Exceptions;
with Ada.Text_IO;                  use Ada.Text_IO;
with Ada.Strings.Fixed;            use Ada.Strings.Fixed;
with Ada.Strings.Unbounded;        use Ada.Strings.Unbounded;

with Unchecked_Deallocation;

--* ASCL Support

package body ASCL.Debugging_Support  is

   Version : constant String := "$Id: ascl-debugging_support.adb,v 1.2 2000/01/23 09:48:34 ACDE Released ACDE $";

   Module  : constant String := "ASCL.Debugging_Support";
   ---====================================================================---
   ---===                O B J  E C T     D A T A                      ===---
   ---====================================================================---

   type Trace_Element;
   type Trace_Element_Access is access Trace_Element;

   type Trace_Element is record
         next : Trace_Element_Access := null;
         item : Unbounded_String;
      end record;

   procedure Free is
      new Unchecked_Deallocation( Trace_Element, Trace_Element_Access);

   --|
   --| This is the instance of the component internal data.
   --|
   type Object_Data is record
         level      : Debug_Level := Quiet;
         call_level : Natural     := 0;
         first      : Trace_Element_Access := null;
      end record;

   ---=====================================================================---
   ---===         L O C A L   S U P P O R T   P R O C E D U R E S       ===---
   ---=====================================================================---

   --|
   --| This procedure is called upon every exception raised in this
   --| component. It may be used to trace back the cause of an
   --| exception, or to print out relevant information about the
   --| instance where the problem did occure.
   --|

   procedure Error( this     : in Object'Class;
                    theError : in Exception_Occurrence;
                    info     : in String ) 
   is
      data : Object_Data_Access := this.data;
   begin
      Put_Line("**** Internal Error ****");
      Put_Line("Exception " &
                Exception_Name( theError ) &
                " in " &
               Module & info );
      Raise_Exception( Exception_Identity( theError ), Module & ":" & info );
   end Error;

   --|
   --| Get pointer to object data
   --|
   function Instance( this : Handle; info : String ) 
                                               return  Object_Data_Access is
   begin
      if this = null then
         return null;
      end if;

      if this.Data = null then
         raise Not_Initialized;
      end if;

      return this.Data;
   exception
      when The_Error: Others => 
         Error( this.all, The_Error, ".Instance in " & info );
	 raise;
   end Instance;


   --|
   --| Trace into some makred section.
   --|
   procedure Trace_Into( data : Object_Data_Access; text : in String ) is
      element : Trace_Element_Access := new Trace_Element;
   begin
      element.item := To_Unbounded_String( text );

      if data.first = null then
         data.first := element;
      else
         element.next := data.first;
         data.first := element;
      end if;

   exception
      when The_Error : Others =>
         Put_Line(Module & ": Internal Error in Trace_Into");
         raise;
   end Trace_Into;

   --|
   --| Return from the last trace point.
   --|
   procedure Return_From( data : in Object_Data_Access ) is
      element : Trace_Element_Access := null;
   begin
      if data.first /= null then
         element := data.first;
         data.first := element.next;
	 Free( element );
      end if;

   exception
      when The_Error : Others =>
         raise;
   end Return_From;

   --|
   --| generate a back trace.
   --|
   procedure Back_Trace( this : in Handle ) is
      data    : Object_Data_Access := this.data;
      element : Trace_Element_Access := data.first;
   begin
      while element /= null loop
         Output( this.all, To_String( element.item ) );
         element := element.next;
      end loop;

   exception
      when The_Error : Others =>
         Error( this.all, The_Error, ".Back_Trace" );
         raise;
   end Back_Trace;		 

   ---======================================================================---
   ---===             C O M P O  N E N T    I N T E R F A C E            ===---
   ---======================================================================---

   --|
   --| Intialize the Object data and call the Initalization
   --| procedure of the extention.
   --|
   procedure Initialize( this  : in out Object'Class ) is
      data             : Object_Data_Access renames this.data;
   begin
      if data /= null then
         raise Usage_Error;
      end if;

      data := new Object_Data;
      if data = null then
         raise Out_Of_Memory;
      end if;

   exception
      when The_Error : Others =>
         Error( this, The_Error, ".Initialize");
   end Initialize;

   --|
   --| Finalize the instance by releasing the object data
   --| finalizing the extention.
   --|
   procedure Finalize( this : in out Object'Class ) is
      data   : Object_Data_Access renames this.data;

      procedure Free is
            new Unchecked_Deallocation( Object_Data, Object_Data_Access);
   begin
      if data = null then
         return;
      end if;

      Free( data );
      this.data := null;

   exception
      when The_Error : Others =>
         Error( this, The_Error, ".Finalize");
   end Finalize;

   ---=====================================================================---
   ---===           A T T R I B U T E    F U N C T I O N S              ===---
   ---=====================================================================---

   --|
   --| This function reads an attribute
   --|
   function Get( this : in Object'Class ) return Debug_Level is
      data : Object_Data_Access := this.data;
   begin
      if data = null then
         raise Not_Initialized;
      end if;

      return data.level;

   exception
      when The_Error : Others =>
         Error( this, The_Error, ".Get Debug_Level");
         raise;
   end Get;

   --|
   --| Set the debuging level.
   --|
   procedure Set( this : in out Object'Class; level : in Debug_Level ) is
      data : Object_Data_Access := this.data;
   begin
      if data = null then
         raise Not_Initialized;
      end if;

      data.level := level;

   exception
      when The_Error : Others =>
         Error( this, The_Error, ".Set Debug_Level" );
         raise;
   end Set;

   ---=====================================================================---
   ---===                        M E T H O D S                          ===---
   ---=====================================================================---

   --|
   --| Purpose:
   --|    Display debugging information upon entering the context
   --|    of a procedure, block etc.... The trace information is
   --|    stored in the trace list.
   --|
   --| Pre-Conditions:
   --|    Instance has been initialized.
   --|
   --| Post-Conditions:
   --|    not applicable
   --|
   procedure Enter( this : in Handle; text : in String ) is
      data       : Object_Data_Access   := Instance( This, ".Enter" );
      element    : Trace_Element_Access := null;
   begin
      if data = null then
         return;
      end if;

      data.call_level := data.call_level + 1;

      case data.level is
         when Logic | Verbose =>
            Output( this.all, data.call_level * "   " & "-> " & text );
         when Others =>
            null;
      end case;

      Trace_Into( data, text );

   exception
      when The_Error : Others =>
         --Back_Trace( this );
         Error( this.all, The_Error, ".Enter" );
         raise;
   end Enter;

   --|
   --| Purpose:
   --|    Display debugging information upon leaving the context
   --|    of a procedure, block etc...
   --|
   --| Pre-Conditions:
   --|    Instance has been initialized.
   --|
   --| Post-Conditions:
   --|    not applicable
   --|
   procedure Leave( this : in Handle; text : in String ) is
      data : Object_Data_Access := Instance( this, ".Leave" );
   begin
      if this = null then
         return;
      end if;

      case data.level is
         when Logic | Verbose  =>
            Output( this.all,
                   data.call_level * "   " &
                   "<- " & To_String( data.first.item ) & " " & text );
          when Others =>
            null;
      end case;

      if data.call_level /= 0 then
         data.call_level := data.call_level - 1;
      end if;

      Return_From( data );

   exception
      when The_Error : Others =>
         Error( this.all, The_Error, ".Enter" );
         raise;
   end Leave;

   --|
   --| Purpose:
   --|    Display debugging information about an error occured in
   --|    the application.
   --|
   --| Pre-Conditions:
   --|    Instance has been initialized.
   --|
   --| Post-Conditions:
   --|    not applicable
   --|
   procedure Error( this : in Handle; text : in String ) is
      data : Object_Data_Access := Instance( this, ".Error");
   begin
      if this = null then
         return;
      end if;

      Output( this.all, text );
      Output( this.all, "Backtrace:");

      Back_Trace( this );

   exception
      when The_Error : Others =>
         Error( this.all, The_Error, ".Enter" );
         raise;
   end Error;


   --|
   --| Purpose:
   --|    Display debugging information
   --|
   --| Pre-Conditions:
   --|    Instance has been initialized.
   --|
   --| Post-Conditions:
   --|    not applicable
   --|
   procedure Trace( this : in Handle; text : in String ) is
      data : Object_Data_Access := Instance( this, ".Trace" );
   begin
      if this = null then
         return;
      end if;

      case data.level is
         when Verbose | Logic =>
            Output( this.all, 
                    data.call_level * "   " & "  " &  text );
         when Others =>
            null;
      end case;

   exception
      when The_Error : Others =>
         Error( this.all, The_Error, ".Trace" );
         raise;
   end Trace;

end ASCL.Debugging_Support;
