--|                              -*- Mode: Ada -*-
--| Filename        : $Source: /home/ACDE/ACDE/ASCL/RCS/ascl-dt-index_tree.adb,v $
--| Description     :
--| Author          : Michael Erdmann
--| Created On      :
--| Last Modified By: $Author: ACDE $
--| Last Modified On: $Date: 2000/01/23 09:48:34 $
--| Update Count    : 0
--| Status          : $State: Released $
--|
--| Functional Description
--| ======================
--| A tree is build from leafes. A leaf contains a reference
--| to the data and a set of references to other leafes.
--| Each item to be inserted is identitified by a key.
--|
--|
--|	     
--| History
--| =======			  
--| $Log: ascl-dt-index_tree.adb,v $
--| Revision 1.2  2000/01/23 09:48:34  ACDE
--| Release Version base_1_1
--|
--| Revision 1.1  2000/01/23 09:46:34  ACDE
--| Release Version 1_0
--|
--| Revision 1.7  1999/11/27 16:56:18  ACDE
--| Release Version 1P010
--|
--|
--| Revision 1.1  1999/11/27 16:39:41  ACDE
--| No comments
--|
--|
--|

--* Ada
with Ada.Exceptions;               use Ada.Exceptions;
with Ada.Streams;                  use Ada.Streams;
with Ada.Streams.Stream_IO;        use Ada.Streams.Stream_IO;
with Ada.Text_IO;                  use Ada;

with Unchecked_Deallocation;

with ASCL.DT.Tree_Element;            use ASCL.DT.Tree_Element;

use ASCL;
use ASCL.DT;

package body ASCL.DT.Index_Tree is

   ---=====================================================================---
   ---===                   L O C A L   D A T A                         ===---
   ---=====================================================================---
   Version : constant String := "$Id: ascl-dt-index_tree.adb,v 1.2 2000/01/23 09:48:34 ACDE Released ACDE $";

   type Leaf;
   type Reference is access Leaf;

   type Reference_Array is array( Natural range <> ) of Reference;

   type Leaf( size : Positive ) is record
         First_Time_Visited  : Boolean                  := False;
         item                : Tree_Element.Handle := null;
         next                : Positive                 := 1;
         parent              : Reference                := null;
         sub_item            : Reference_Array(1..size) := (others=>null);
      end record;

   ---====================================================================---
   ---===                O B J  E C T     D A T A                      ===---
   ---====================================================================---

   type Object_Data is record
         width       : Integer   := 0;
         root        : Reference := null;
         current     : Reference := null;
      end record;


   ---=====================================================================---
   ---===         L O C A L   S U P P O R T   P R O C E D U R E S       ===---
   ---=====================================================================---

   procedure Error(  theError : in Exception_Id ; info : in String ) is
   begin
      Text_IO.Put_Line("*** ascl-tree.adb" & info );
      Raise_Exception( theError,  "ascl-tree.adb" & info );
   end Error;

   --|
   --| Return an initialized version of an object instance.
   --|
   function Initialize ( size : Integer ) return Object_Data_Access is
      result : Object_Data_Access := new Object_Data;
   begin
      result.width := size;

      return result;
   end Initialize;

   ---======================================================================---
   ---===             C O M P O  N E N T    I N T E R F A C E            ===---
   ---======================================================================---

   --|
   --| Intialize the Object data and call the Initalization
   --| procedure of the extention.
   --|
   procedure Initialize( this : in out Object'Class ) is
      data             : Object_Data_Access renames this.data;
   begin
      if data = null then
         raise Out_Of_Memory;
      end if;

   exception
      when The_Error : Others =>
         Error( Exception_Identity( The_Error ), ".Initialize");
         raise;
   end Initialize;

   --|
   --| Finalize the instance by releasing the object data
   --| finalizing the extention.
   --|
   procedure Finalize( this : in out Object'Class ) is
      data   : Object_Data_Access renames this.data;

      procedure Free is
            new Unchecked_Deallocation( Object_Data, Object_Data_Access);
   begin
      if data = null then
         return;
      end if;

      Free( data );

      this.data := null;
   end Finalize;

   ---=====================================================================---
   ---===           A T T R I B U T E    F U N C T I O N S              ===---
   ---=====================================================================---

   --|
   --| This function reads the width of a single key
   --| value.
   --|
   function Width( this : in Object'Class ) return Integer is
      data : Object_Data_Access := this.data;
   begin
      if data = null then
         raise Not_Initialized;
      end if;

      return data.width;

   exception
      when The_Error : Others =>
         Error( Exception_Identity( The_Error ), ".Width");
         raise;

   end width;


   ---=====================================================================---
   ---===                        M E T H O D S                          ===---
   ---=====================================================================---

   --|
   --|  Insert a tree element into a tree.
   --|
   procedure Add( this : in out Object'Class;
                  item : in Tree_Element.Handle )
   is
      data     : Object_Data_Access renames this.data;
      Current  : Reference := null;
   begin
      if data = null then
         raise Not_Initialized;
      end if;

      if data.root = null then
         data.root := new Leaf( data.width );
      end if;
      Current := data.root;

      for Position in 1..Length( item.all ) loop
         declare
            idx : Integer := Key( item.all, Position );
         begin
            if current.sub_item( idx ) = null then
               current.sub_item( idx ) := new Leaf( data.width );
               current.sub_item( idx ).parent := current;
            end if;

            current := current.sub_item( idx );
         end;
      end loop;
      current.item := item;

   exception
      when The_Error : Others =>
         Error( Exception_Identity( The_Error ), ".Add");
         raise;
   end Add;

   --|
   --| This procedure walks through the tree and executes
   --| a procedure for every element found.
   --|
   procedure Process( this : in out Object'Class ) is
      data     : Object_Data_Access := this.data;
      level    : Integer := 0;

      procedure Walk_Leaf( leaf : in Reference ) is
      begin
         level := level + 1;
         if leaf.item /= null then
            Action( this, leaf.item, level );
         end if;

         for i in 1..data.width loop
            if leaf.sub_Item(i) /= null then
               Walk_Leaf( leaf.sub_item(i) );
            end if;
         end loop;
         level := level - 1;

      exception
         when The_Error : Others =>
            Error( Exception_Identity( The_Error ), ".Walk_Leaf");
            raise;
      end Walk_Leaf;

   begin
      if data = null then
         raise Not_Initialized;
      end if;

      if data.root = null then
         return;
      else
         Walk_Leaf( data.root );
      end if;

   exception
      when The_Error : Others =>
         Error( Exception_Identity( The_Error ), ".Process" );
   end Process;

   --|
   --| This procedure provides an sequential access to the
   --| index tree.
   --|
   --|
   function Get( this : in Object;
                 mode : Selection := Next )  return Tree_Element.Handle is
      data    : Object_Data_Access  := this.data;
      Result  : Tree_Element.Handle := null;
      current : Reference := null;
      next    : Reference := null;

      -- find the first non null reference
      function Next_Entry( current : in Reference ) return Reference is
         Result  : Reference := null;
      begin
         for i in current.next..current.sub_item'length loop
            if current.sub_Item(i) /= null then
               Result := current.sub_Item(i);
               current.next := i + 1;
               exit;
            end if;
         end loop;

         if Result = null then
            current.next := 1;
         end if;

         return Result;

      exception
         when The_Error : Others =>
            Error( Exception_Identity( The_Error ), "Get.Next");
            raise;
      end Next_Entry;

   begin

      if mode = First  then
         data.current := data.root;
      end if;

      current := data.current;

      Search_For_Item:
         while Result = null loop

            if not current.First_Time_Visited then
               current.First_Time_Visited := True;
            else
               next := Next_Entry( current );
               while next = null loop
                  current.First_Time_Visited := False;

                  current := current.parent;

                  if current = null then
                     exit Search_For_Item;
                  end if;
                  next := Next_Entry( current );
               end loop;
               current := next;
            end if;

            Result := current.Item;

         end loop Search_For_Item;

      data.current := current;

      return Result;

   exception
      when The_Error : Others =>
         Error( Exception_Identity( The_Error ), ".Get");
         raise;
   end Get;

   --|
   --| This procedure saves the contents information into
   --| an file stream.
   --|
   procedure Save( this : in out Object'Class; fileName : in String ) is
      data          : Object_Data_Access  := this.data;
      item          : Tree_Element.Handle := null;

      Target_File   : File_Type;
      outFile       : Stream_Access;
   begin
      if data = null then
         raise Not_Initialized;
      end if;

      Create( File => Target_File, mode => Out_File, Name => fileName );
      outFile := Stream( Target_File );

      item := Get( this, mode => First );
      while item /= null loop

         Write( outFile, item.all );
         item := Get( this );
      end loop;

      Close( Target_File );

   exception
      when The_Error : Others =>
         Error( Exception_Identity( The_Error ), ".Save" );
         raise;
   end Save;

   --|
   --| Load the tree from an external file
   --|
   procedure Load( this : in out Object'Class; fileName : in String ) is
      data          : Object_Data_Access  := this.data;
      item          : Tree_Element.Handle := null;

      Input_File    : File_Type;
      inStream      : Stream_Access;
   begin

      Open( File => Input_File, mode => In_File, Name => fileName );
      inStream := Stream( Input_File );

      while not End_Of_File( Input_File ) loop
         null;
      end loop;

      Close( Input_File );


   exception
      when The_Error : Others =>
         raise;
   end Load;
end ASCL.DT.Index_Tree;

