--|
--| Filename        : $Source: /home/ACDE/ACDE/ASCL/RCS/ascl-dt-list.adb,v $
--| Description     : User Object
--| Author          : Michael Erdmann
--| Created On      : 25.3.1999
--| Last Modified By: $Author: ACDE $
--| Last Modified On: $Date: 2000/01/23 09:48:34 $
--| Status          : $State: Released $
--|
--| Functional Description
--| ======================
--| This package implements a simple list of elements of the 
--| same type.
--|
--| Restrictions
--| ============
--| None
--|
--| References
--| ==========
--|
--|
--| History
--| =======
--| $Log: ascl-dt-list.adb,v $
--| Revision 1.2  2000/01/23 09:48:34  ACDE
--| Release Version base_1_1
--|
--| Revision 1.1  2000/01/23 09:46:34  ACDE
--| Release Version 1_0
--|
--| Revision 1.13  1999/11/28 12:20:33  ACDE
--| Instance Function added
--|
--|

--* Ada
with Ada.Exceptions;               use Ada.Exceptions;
with Unchecked_Deallocation;

with ASCL.Debugging_Support;  	   use ASCL.Debugging_Support;
use  ASCL;

package body ASCL.DT.List is

   Version : constant String := "$Id: ascl-dt-list.adb,v 1.2 2000/01/23 09:48:34 ACDE Released ACDE $";
   Module  : constant String := "ascl-dt-list.adb";

   ---====================================================================---
   ---===                O B J  E C T     D A T A                      ===---
   ---====================================================================---

   type List_Element;
   type List_Element_Access is access List_Element;

   type List_Element is record
         Next     : List_Element_Access := null;
	 Previous : List_Element_Access := null;

	 Data     : Item_Type ;
      end record;

   --|
   --| This is the instance of the component internal data.
   --|
   type Object_Data is record
         Debug   : Debugging_Support.Handle := null;

	 Head    : List_Element_Access := null;
	 Tail    : List_Element_Access := null;
      end record;

   ---=====================================================================---
   ---===         L O C A L   S U P P O R T   P R O C E D U R E S       ===---
   ---=====================================================================---

   --|
   --| This procedure is called upon every exception raised in this
   --| component. It may be used to trace back the cause of an
   --| exception, or to print out relevant information about the
   --| instance where the problem did occure.
   --|
   procedure Error( this     : in Object'Class;
                    theError : in Exception_Occurrence;
                    info     : in String ) 
   is
      data : Object_Data_Access := this.data;
   begin
      Trace( data.Debug, "*** Exception : " & Exception_Name( theError ) );
      Trace( data.Debug, "              : " & Exception_Message( theError) );
      Leave( data.Debug, "*** Exception ***");

      Raise_Exception( Exception_Identity( theError ),  Module & info );
   end Error;

   --|
   --| Get pointer to object data
   --|
   function Instance( this : Object'Class; info : String ) 
                                               return  Object_Data_Access is
   begin
      if this.Data = null then
         raise Not_Initialized;
      end if;

      return this.Data;
   exception
      when The_Error: Others => 
         Error( this, The_Error, ".Instance");
	 raise;
   end Instance;


   ---======================================================================---
   ---===             C O M P O  N E N T    I N T E R F A C E            ===---
   ---======================================================================---

   --|
   --| Intialize the Object data and call the Initalization
   --| procedure of the extention.
   --|
   procedure Initialize( this  : in out Object'Class;
                         debug : in Debugging_Support.Handle := null ) 
   is
      Data  : Object_Data_Access renames this.data;
   begin
      Enter( debug, Module & ".Initialize" );
      if Data /= null then
         raise Usage_Error;
      else
         Data := new Object_Data;
         if Data = null then
            raise Out_Of_Memory;
         end if;
      end if;

      Data.Debug := debug; 
      
      Leave( debug, "");
   exception
      when The_Error : Others =>
         Error( this, The_Error, ".Initialize");
   end Initialize;

   --|
   --| Finalize the instance by releasing the object data
   --| finalizing the extention.
   --|
   procedure Finalize( this : in out Object'Class ) is
      data   : Object_Data_Access renames this.data;

      procedure Free is
            new Unchecked_Deallocation( Object_Data, Object_Data_Access);
   begin
      if data = null then
         return;
      end if;

      Free( data );

      this.data := null;

   exception
      when The_Error : Others =>
         Error( this, The_Error, ".Finalize");
   end Finalize;

   ---=====================================================================---
   ---===           A T T R I B U T E    F U N C T I O N S              ===---
   ---=====================================================================---


   ---=====================================================================---
   ---===                        M E T H O D S                          ===---
   ---=====================================================================---

   --|
   --| Purpose:
   --|    Append an element to a list
   --|
   --| Pre-Conditions:
   --|    List has to be initialized.
   --|
   --| Post-Conditions:
   --|	  none
   --|
   procedure Append( this : in out Object'Class; item : in Item_Type ) is
      Data   : Object_Data_Access := Instance( this, ".Append");
      Debug  : Debugging_Support.Handle renames Data.Debug;
      Element: List_Element_Access := null;
   begin
      Enter( Debug, Module & ".Append" );

      Element := new List_Element;
      Element.Data := item;

      if Data.Head = null then
         Data.Head := new List_Element;
      else
         Data.Tail.Next := Element;  
      end if;

      Element.Previous := Data.Tail;
      Data.Tail := Element;

      Leave( Debug, "");
   exception
      when The_Error : Others =>
         Error( this, The_Error, ".Append" );
         raise;
   end Append;

   --|
   --| Purpose:
   --|    Perform an action upon the list. The is no update 
   --|    of the elements possible.
   --|
   --| Pre-Conditions:
   --|    Object has to be initialized.
   --|
   --| Post-Conditions:
   --|
   --|
   procedure Perform( this : in out Object'Class; handler : in List_Handler ) is
      data   : Object_Data_Access := Instance( this, ".Perform");
      Debug  : Debugging_Support.Handle renames Data.Debug;
      Element: List_Element_Access := null;
   begin
      Enter( Debug, Module & ".Perform");

      Element := Data.Head;
      while Element /= null loop
         exit when Handler( this, Element.Data ) = False;

	 Element := Element.Next;
      end loop;

      Leave( Debug, "" );

   exception
      when The_Error : Others =>
         Error( this, The_Error, ".Perform" );
         raise;
   end Perform;


end ASCL.DT.List;

