--                              -*- Mode: Ada -*-
-- Filename        : $Source: /home/ACDE/ACDE/ASCL/RCS/ascl-ob-memory_pool.adb,v $
-- Description     : Object Memory Store
-- Author          : Michael Erdmann
-- Created On      : Sun Jul 26 08:24:57 1998
-- Last Modified By: $Author: ACDE $
-- Last Modified On: $Date: 2000/01/23 09:48:34 $
-- Status          : $State: Released $
--
-- Functional Description
-- ======================
-- This package implements an simple memory pool where the
-- components are linked into a single list. It is poosible
-- to store the object data in a file and reload it into
-- the memory.
--
--
-- History
-- =======
-- $Log: ascl-ob-memory_pool.adb,v $
-- Revision 1.2  2000/01/23 09:48:34  ACDE
-- Release Version base_1_1
--
-- Revision 1.1  2000/01/23 09:46:34  ACDE
-- Release Version 1_0
--
-- Revision 1.11  1999/11/27 16:56:18  ACDE
-- Release Version 1P010
--
-- Revision 1.1  1999/11/27 09:33:55  ACDE
-- sg
--
--
with System;                   use System;
with System.Storage_Elements;  use System.Storage_Elements;

with Ada.Exceptions;           use Ada.Exceptions;
with Ada.Streams.Stream_IO;    use Ada.Streams.Stream_IO;
with Ada.Strings;              use Ada.Strings;
with Ada.Strings.Fixed;        use Ada.Strings.Fixed;

with Unchecked_Deallocation;

with ASCL.Debugging_Support;   use ASCL.Debugging_Support;

use  ASCL;


package body ASCL.OB.Memory_Pool is

   Version : constant String := "$id";
   Module  : constant String := "ASCL.OB.Memory_Pool";

   ---======================================================================---
   ---====             C O M P O N E N T   D A T A                       ===---
   ---======================================================================---

   --|
   --| This data structure contains the data of an OMS-Object,
   --| which is basicaly the head and tail pointer of the
   --| memory chunk list and the object table it self.
   --|
   --| The object table allows a mapping between the object identifier
   --| and the adfa object it self.
   --|
   type Chunk_List_Element;
   type Chunk_List_Element_Ptr is access Chunk_List_Element;

   type Object_Data is record
         File_Name    : String(1..250);
         Used_Memory  : Storage_Count := 0;
         Max_Used     : Storage_Count := 64_000;

         first        : Chunk_List_Element_Ptr   := null;
         last         : Chunk_List_Element_Ptr   := null;
         debug        : Debugging_Support.Handle := null;
      end record;

   --|
   --| A memory chunk is a block of storage elements which is
   --| used to store Ada objects. The memory chunk is part
   --| of the memory chunk list. Such an element is defined by
   --| the chunk list element.
   --|
   type Chunk_List_Element( Size : Storage_Count )  is record
          ID       : Integer;
          next     : Chunk_List_Element_Ptr := null;
          memory   : Storage_Array( 1..Size) ;
       end record;

   ---======================================================================---
   ---====        I N T E R N A L   P R O C E D U R E S                  ===---
   ---======================================================================---


   --|
   --| This procedure is called upon every error found (exception)
   --| in this package.
   --|
   procedure Error( this     : in Object;
                    theError : in Exception_Id ;
                    info     : in String )
   is
      data : Object_Data_Access := this.data;
   begin
      Error( data.debug,
             "*** Exception " & Exception_Name( theError ) &
                " in " & info );
      Raise_Exception( theError,  Module & ":" & info );
   end Error;

   --|
   --| Retrieve the pointer to the object data. If the object is not
   --| yet initialized raise an exception.
   --|
   function GetData( this : in Object; where : String  ) return Object_Data_Access is
   begin
      if this.Data = null then
         Error( this, Not_Initialized'Identity,
                      "(oms.adb): GetData - Object not defined in :" & where );
      end if;

      return this.Data;
   end GetData;

   ---======================================================================---
   ---====         C O M P O N E N T   I N T E R F A C E                 ===---
   ---======================================================================---

   --|
   --| Initialize the object data
   --|
   procedure Initialize( this : in out Object'Class;
                         debug: in Debugging_Support.Handle := null )
   is
      new_data : Object_Data_Access := null;
   begin
      if this.data /= null then
         Raise_Exception( Usage_Error'Identity,
                        "(oms.adb): procedure Initialize");
      else
         new_data := new Object_Data;
         this.data := new_data;

         new_data.debug := debug;
      end if;
   end Initialize;

   --|
   --| Finalize the object by deallocating the storage pool
   --|
   procedure Finalize( this : in out Object ) is
      data : Object_Data_Access := this.data;
   begin
      null;
   end Finalize;

   ---======================================================================---
   ---===           A T T R I B U T E   F U N C T I O N S                ===---
   ---======================================================================---

   --|
   --|  Return the current filename
   --|
   function File_Name (this : in Object) return String is
      data : Object_Data_Access := GetData( this, "Get File_Name" );
   begin
      return  Trim( data.File_Name, Side => Right );
   end File_Name;

   --|
   --| Set the filename for the save file
   --|
   procedure File_Name( this : in out Object; name : in String ) is
      data : Object_Data_Access := GetData( this, "Set File_Name");
   begin
      Move( name, data.File_Name );
   end File_Name;


   ---======================================================================---
   ---===                  O M S    F U N C T I O N S                    ===---
   ---======================================================================---

   --|
   --| Store the contents of the object pool in the file specified in
   --| the oms object data.
   --|
   procedure Store( this : in Object ) is
      data : Object_Data_access := GetData( this, "Store" );
      file : File_Type;
      outp : Stream_Access := null;
      chunk: Chunk_List_Element_Ptr := data.first;
   begin
      Enter( data.debug, "Store");

      Create( file,
              Mode => Out_File,
              Name => File_Name( this ) );

      outp := Stream( file );

      while Chunk /= null loop
         Integer'Write( outp, Chunk.ID );

         Integer'Write( outp, Chunk.memory'Size/Storage_Unit);
         Storage_Array'Write( outp, Chunk.memory );

         Chunk := Chunk.next;
      end loop;

      Integer'Write( outp, 0 );

      Close(file);
      Leave( data.debug, "");

   exception
      when Error : others =>
         Raise_Exception( Exception_Identity(Error),
                        "(oms.adb): Store of file " & File_Name(this) );
   end Store;

   --|
   --| Load oms data from a file.
   --|
   procedure Load( this : in Object ) is
      data : Object_Data_access := GetData( this, "Store" );
      file : File_Type;
      inp  : Stream_Access := null;
   begin

      Enter( data.debug, "Load");

      Open( file, Mode => In_File, Name => File_Name(this));
      inp := Stream( file );

      loop
         declare
            id     : Integer ;
            length : Storage_Count ;
            chunk  : Chunk_List_Element_Ptr := null;
         begin

            Integer'Read(inp, id );
            if id = 0 then
               exit;
            end if;

            Storage_Count'Read(inp, length  );
            chunk := new Chunk_List_Element( length );
            chunk.next := null;

            if data.first = null then
               data.first := chunk;
            else
               data.last.next := chunk;
            end if;
            data.last := chunk;
            Storage_Array'Read(inp, chunk.memory );
         end;
      end loop;

      Close( file );

      Leave( data.debug, "");

   exception
      when The_Error : others =>
         Error( this, Exception_Identity( The_Error ), ".Load"  );
   end Load;

   ---======================================================================---
   ---===          PROCEDRUE REQUIRED BY MEMORY POOL MANAGEMENT          ===---
   ---======================================================================---


   --|
   --| Allocate a new object. Return the address of the object
   --| table entry and allocate the object based upon the
   --| class object assigned to the pool
   --|
   procedure Allocate( Pool            : in out Object;
                       Storage_Address : out    Address;
                       Size            : in     Storage_Count;
                       Alignment       : in     Storage_Count )
   is
      data     : Object_Data_Access := GetData( Pool, "Allocate" );
      chunk    : Chunk_List_Element_Ptr := null;
   begin
      Enter( data.debug, "Allocate" );

      if data.used_memory + Size > data.max_used then
         raise Pool_Exceeded;
      end if;

      chunk := new Chunk_List_Element(Size);
      if chunk = null  then
         raise Memory_Allocation_Failure;
      end if;
      chunk.next := null;

      if data.first = null then
         data.first := chunk;
      else
         data.last.next := chunk;
      end if;
      data.last := chunk;

      Storage_Address  := chunk.memory'Address;
      data.used_memory := data.used_memory + Size;

   exception
      when The_Error : Others =>
         Error( Pool, Exception_Identity( The_Error ), ".Allocate");
   end Allocate;

   --|
   --| Deallocate memory.
   --|
   procedure Deallocate( Pool          : in out Object;
                       Storage_Address : in     Address;
                       Size            : in     Storage_Count;
                       Alignment       : in     Storage_Count )   is
      data : Object_Data_Access := GetData( pool, "Deallocate" );

      procedure Free is new Unchecked_Deallocation(Chunk_List_Element,
                                         Chunk_List_Element_Ptr);
      Chunk : Chunk_List_Element_Ptr := data.first;
      Prev  : Chunk_List_Element_Ptr := null;
   begin

      Enter( data.debug, "Deallocate");

      while Chunk /= null loop
         if Chunk.memory'Address = Storage_Address then
            if Prev = null then
               data.first := Chunk.next;
            else
               Prev.next := Chunk.next;
            end if;

            data.used_memory := data.used_memory - Size;
            exit;
         else
            Prev := Chunk;
         end if;
      end loop;

      Leave( data.debug, "");

   exception
      when The_Error : Others =>
         Error( pool, Exception_Identity(The_Error), ".Deallocate");
   end Deallocate;

   --|
   --| Return the size of the memory pool. This value may change during
   --| the lifetime of an object pool.
   --|
   function Storage_Size ( pool : Object ) return Storage_Count is
      Data   : Object_Data_Access := GetData( pool, "Storage_Size" );
   begin
      return Data.Max_Used;
   end Storage_Size;

end ASCL.OB.Memory_Pool;

