--|        
--| Filename        : $Source: /home/ACDE/ACDE/ASCL/RCS/ascl-rt-fsm.adb,v $
--| Description     : Finit State Machine framework
--| Author          : Michael Erdmann
--| Created On      : 25.3.1999
--| Last Modified By: $Author: ACDE $
--| Last Modified On: $Date: 2000/01/23 09:48:34 $
--| Status          : $State: Released $
--|
--| Functional Description
--| ======================
--|
--|
--|
--| History
--| =======
--| $Log: ascl-rt-fsm.adb,v $
--| Revision 1.2  2000/01/23 09:48:34  ACDE
--| Release Version base_1_1
--|
--| Revision 1.1  2000/01/23 09:46:34  ACDE
--| Release Version 1_0
--|
--| Revision 1.11  1999/11/27 16:56:18  ACDE
--| Release Version 1P010
--|
--| Revision 1.1  1999/11/27 09:33:55  ACDE
--| sg
--|
--| Revision 1.1  1999/07/26 03:56:56  erdmann
--| Initial Version for testing
--|

--* Ada
with Ada.Exceptions;               use Ada.Exceptions;
with Unchecked_Deallocation;

with ASCL.RT.Event;
with ASCL.RT.Event_Q;      use ASCL.RT.Event_Q;
with ASCL.RT.State;        use ASCL.RT.State;

use ASCL.RT;

package body ASCL.RT.Fsm is

   Version : constant String := "$Id: ascl-rt-fsm.adb,v 1.2 2000/01/23 09:48:34 ACDE Released ACDE $";

   ---====================================================================---
   ---===                O B J  E C T     D A T A                      ===---
   ---====================================================================---

   --|
   --| This is the instance of the component internal data.
   --|
   type State_Table is array( State.State_ID ) of State.Handle;

   task type Supervisor( this : Object_Data_Access );
   task type FSM( this : Object_Data_Access );

   type Supervisor_Access is access Supervisor;
   type FSM_Access is access FSM;

   type Object_Data is record
         Input         : Event_Q.Object;
         FSM_Task      : FSM_Access        := null;
         Sup_Task      : Supervisor_Access := null;
         Nbr_Of_Events : Natural           := 0;
         States        : State_Table       := (others => null );
         Initial       : State.State_ID    := 0;
         Final         : State.State_ID    := 0;
      end record;


   ---=====================================================================---
   ---===         L O C A L   S U P P O R T   P R O C E D U R E S       ===---
   ---=====================================================================---

   --|
   --| This procedure is called upon every exception raised in this
   --| component. It may be used to trace back the cause of an
   --| exception, or to print out relevant information about the
   --| instance where the problem did occure.
   --|
   procedure Error(  theError : in Exception_Id ;
                     info     : in String;
                     data     : in Object_Data_Access ) is
   begin
      Raise_Exception( theError,  "ascl-rt-fsm.adb" & info );
   end Error;

   --|
   --| This is the supervisor task. Every second a check for
   --| activity is done. If no activity has been seen the
   --| prcoess will be aborted.
   --|
   task body Supervisor is
      count : Natural := 0;
   begin
      loop
         this.Nbr_Of_Events := 0;
         delay 1.0;
         if this.nbr_Of_Events = 0 then
            abort this.FSM_Task.all;
         end if;
      end loop;

   end Supervisor;

   --|
   --| This is the fsm task. It is monitored by the Supervisor
   --| for activitiy.
   --|
   task body FSM is
      Next  : State.Handle   := null;
      id    : State.State_ID := State.Final;
      Evt   : Event.Handle   := null;
   begin
      Next := This.States( this.Initial );

      loop
         Evt := Get( This.Input );
         This.Nbr_Of_Events := this.Nbr_Of_Events + 1;

         Id := Respond( Next.all, Evt );
         if Id = This.Final then
            exit;
         else
            Next := This.States( id );
         end if;

         if Next = null then
            raise Usage_Error;
         end if;
      end loop;

      abort This.Sup_Task.all;
   end FSM;

   ---======================================================================---
   ---===             C O M P O  N E N T    I N T E R F A C E            ===---
   ---======================================================================---

   --|
   --| Intialize the Object data and call the Initalization
   --| procedure of the extention.
   --|
   procedure Initialize( this  : in out Object'Class;
                         debug : in Debugging_Support.Handle := null ) 
   is
      data : Object_Data_Access renames this.data;
   begin
      if data /= null then
         raise Usage_Error;
      end if;

      data := new Object_Data;
      if data = null then
         raise Out_Of_Memory;
      end if;

      --* setup the component data
      data.Sup_Task := new Supervisor( data );
      data.FSM_Task := new FSM( data );

   exception
      when The_Error : Others =>
         Error( Exception_Identity( The_Error ), ".Initialize", data);
         raise;
   end Initialize;

   --|
   --| Finalize the instance by releasing the object data
   --| finalizing the extention.
   --|
   procedure Finalize( this : in out Object'Class ) is
      data   : Object_Data_Access renames this.data;

      procedure Free is
            new Unchecked_Deallocation( Object_Data, Object_Data_Access);
   begin
      if data = null then
         return;
      end if;

      Free( data );
      this.data := null;
   end Finalize;

   ---=====================================================================---
   ---===           A T T R I B U T E    F U N C T I O N S              ===---
   ---=====================================================================---

   --|
   --| Register a state with the fsm.
   --|
   procedure Register( this    : in out Object'Class;
                       aState  : in State.Handle;
                       stateID : in State.State_ID ) is
      data : Object_Data_Access renames this.data;
   begin
      if data = null then
         raise Usage_Error;
      end if;

      data.States( stateID ) := aState;

   exception
      when The_Error : Others =>
         Error( Exception_Identity( The_Error ), ".Register", data );
   end Register;

   --|
   --| Set the final state of the fsl.
   --|
   procedure Final( this    : in out Object'Class;
                    aState  : in State.State_ID )
   is
      data : Object_Data_Access renames this.data;
   begin
      if data = null then
         raise Usage_Error;
      end if;

      data.Final := aState;

   exception
      when The_Error : Others =>
         Error( Exception_Identity( The_Error ), ".Final", data );
   end Final;

   --|
   --| Set the intial state
   --|
   procedure Initial( this   : in out Object'Class; aState : in State.State_ID )
   is
      data : Object_Data_Access renames this.data;
   begin
      if data = null then
         raise Usage_Error;
      end if;

      data.Initial := aState;

   exception
      when The_Error : Others =>
         Error( Exception_Identity( The_Error ), ".Final", data );
   end Initial;


   ---=====================================================================---
   ---===                        M E T H O D S                          ===---
   ---=====================================================================---


end ASCL.RT.Fsm;

