--|
--| Filename        : $Source: /home/ACDE/ACDE/ASCL/RCS/ascl-rt-resource.adb,v $
--| Description     : Watched resources.
--| Author          : Michael Erdmann
--| Created On      : 25.3.1999
--| Last Modified By: $Author: ACDE $
--| Last Modified On: $Date: 2000/01/23 09:48:34 $
--| Status          : $State: Released $
--|
--| Functional Description
--| ======================
--|
--| This module maintains for all initialized instances of the
--| resources a list.
--|
--|
--| Restrictions
--| ============
--|
--| The debugging support does not work for the class operations like
--| Register.
--|
--| References
--| ==========
--|
--| History
--| =======
--| $Log: ascl-rt-resource.adb,v $
--| Revision 1.2  2000/01/23 09:48:34  ACDE
--| Release Version base_1_1
--|
--| Revision 1.1  2000/01/23 09:46:34  ACDE
--| Release Version 1_0
--|
--| Revision 1.11  1999/11/27 16:56:18  ACDE
--| Release Version 1P010
--|
--| Revision 1.1  1999/11/27 09:33:55  ACDE
--| sg
--|
--| Revision 1.4  1999/08/01 19:13:35  erdmann
--| Added debugging support (ASCL.Debugging_Support)
--|
--| Revision 1.3  1999/07/26 17:41:39  erdmann
--| Testing complete, see test-resource, main
--|
--| Revision 1.2  1999/07/25 18:12:21  erdmann
--| Coding complete. Ready for test
--|
--|

--* Ada
with Ada.Text_IO;                  use Ada.Text_IO;
with Ada.Tags;                     use Ada.Tags;
with Ada.Strings.Unbounded;        use Ada.Strings.Unbounded;
with Ada.Exceptions;               use Ada.Exceptions;

with Unchecked_Deallocation;

with ASCL.Debugging_Support;       use ASCL.Debugging_Support;
use  ASCL;

package body ASCL.RT.Resource is

   Version : constant String := "$Id: ascl-rt-resource.adb,v 1.2 2000/01/23 09:48:34 ACDE Released ACDE $";
   Module  : constant String := "ASCL.RT.Resource";

   ---====================================================================---
   ---===                  C L A S S   D A T A                         ===---
   ---====================================================================---

   type Class_Record is record
         Name    : Tag;
         Handler : Resource_Handler := null;
      end record;

   type Class_ID is new Integer range 0..250;
   Null_Class_ID : constant Class_ID := 0;

   type Class_Table is array( 1..Class_ID'Last ) of Class_Record;

   task type Watch_Dog_Task;
   type Watch_Dog_Task_Access is access Watch_Dog_Task;

   type Watch_Record is record
         Used       : Boolean                  := False;
         life_time  : Natural                  := 0;
         Handler    : Resource_Handler         := null;
         class_ref  : Class_ID                 := Null_Class_ID;
         debug      : Debugging_Support.Handle := null;
      end record;

   type Watch_Table is array( 1..Resource_ID'Last ) of Watch_Record;

   ---------------------------------------------------------------------------

   WT         : Watch_Table;                      

   Supervisor : Watch_Dog_Task_Access   := null;

   CT         : Class_Table;

   Cls_Debug  : Debugging_Support.Handle := null;

   ---=====================================================================---
   ---===         L O C A L   S U P P O R T   P R O C E D U R E S       ===---
   ---=====================================================================---

   --|
   --| This procedure is called upon every exception raised in this
   --| component. It may be used to trace back the cause of an
   --| exception, or to print out relevant information about the
   --| instance where the problem did occure.
   --|
   procedure Error( this     : in Object'Class;
                    theError : in Exception_Id ;
                    info     : in String ) is
      debug  : Debugging_Support.Handle renames WT(this.id).debug;
   begin
      Error( debug,
             "*** Exception " & Exception_Name( theError ) &
             " in " & Module & info );
      Leave( debug, "");
      Raise_Exception( theError, Module & info );
   end Error;


   --|
   --| This is the supervisor task. It scans every 10 seconds the
   --| resource list and checks wether a resource is still in
   --| use.
   --|
   task body Watch_Dog_Task is
      Used_Entries : Natural := 0;
   begin
      Main : loop
         delay 1.0;

         Used_Entries := 0;
         for i in 1..Resource_ID'Last loop
            if WT(i).Used then
               Used_Entries := Used_Entries + 1;
               if WT(i).life_time /= 0 then

                  WT(i).life_time := WT(i).life_time - 1;
                  if WT(i).life_time = 0 then
                     CT(WT(i).class_ref).Handler( i );
                  end if;

               end if;
            end if;
         end loop;

         exit when Used_Entries = 0;

      end loop Main;

   end Watch_Dog_Task;

   ---======================================================================---
   ---===             C O M P O  N E N T    I N T E R F A C E            ===---
   ---======================================================================---

   --|
   --| Initialize an watched resource by locating the class of
   --| the object in the class table and assigning the debug
   --| object.
   --|
   procedure Initialize( this  : in out Object'Class;
                         debug : Debugging_Support.Handle := null ) is
   begin
      Enter( debug, Module & ".Initialize");

      for i in 1..Class_ID'Last loop
         if CT(i).Handler /= null and then  CT(i).Name = this'Tag then
            WT( this.id ).class_ref := i;
            exit;
         end if;
      end loop;

      WT( this.id ).debug := debug;

      Leave( debug, "");
   exception
      when The_Error : Others =>
         Leave( debug, "exception raised");
         Error( this, Exception_Identity( The_Error ), ".Initialize" );
         raise;
   end Initialize;

   --|
   --| Finalize a watched resource. Currently nothing happens except
   --| for some debugging information.
   --|
   procedure Finalize( this : in out Object'Class ) is
   begin
      null;

   exception
      when The_Error : Others =>
         raise;
   end Finalize;

   --|
   --| Initialize the component 
   --|
   procedure Initialize( debug : in Debugging_Support.Handle := null ) is
   begin
      Cls_Debug := debug;

   exception
      when The_Error : Others =>
         raise;         
   end Initialize;

   --|
   --| Finalize the component by shutting down used resources.
   --|
   procedure Finalize is
   begin
      Cls_Debug := null;
      if Supervisor /= null then
         abort Supervisor.all;
      end if;

   exception
      when The_Error : Others =>
         raise;         
   end Finalize;
   ---===================================================================---
   ---===       C O N T R O L L E D   T Y P E  I N T E R F A C E      ===---
   ---===================================================================---

   --|
   --| This procedure is called by the Controlled type every time
   --| an instance of the resource is created.
   --|
   procedure Initialize( this : in out Watch_Dog ) is
   begin
      if Supervisor = null then
         Supervisor := new Watch_Dog_Task;
      end if;

      for i in 1..Resource_ID'Last loop

         if not WT(i).Used  then
            WT(i).used      := True;
            WT(i).life_time := this.super.life_time;

            this.super.id   := i;
            return;
         end if;
      end loop;

      raise Out_Of_Memory;

   exception
      when The_Error : Others =>
         Error( this.super.all, Exception_Identity(The_Error), ".Initialize");
   end Initialize;

   --|
   --| This procedure in invoked by the controlled
   --| type.
   --|
   procedure Finalize( this : in out Watch_Dog ) is
   begin
      Deallocate( this.super.all );

      WT( this.super.id ).Used := False;
   exception
      when The_Error : Others =>
         Error( this.super.all, Exception_Identity(The_Error),".Finalize");
   end Finalize;


   ---=====================================================================---
   ---===           A T T R I B U T E    F U N C T I O N S              ===---
   ---=====================================================================---

   --|
   --| Return the resource identifier
   --|
   function Id( this : in Object ) return Resource_ID is
      debug : Debugging_Support.Handle renames WT( this.id ).debug;
   begin
      Enter( debug, Module & ".Id");
      if not WT( this.id ).Used then
         raise Usage_Error;
      end if;

      Leave( debug, " yields " & Resource_ID'Image(this.id) );
      return this.id;

   exception
      when The_Error : Others =>
         Error( this, Exception_Identity(The_Error),".function Id");
         return Null_Resource_ID;
   end Id;

   --|
   --| Register the resource handler for the resource in the
   --| class table.
   --|
   procedure Register(  name : Tag ; handler : in Resource_Handler ) is
   begin
      for i in 1..Class_ID'Last loop
         if CT(i).Handler = null then
            CT(i).Name    := name;
            CT(i).Handler := handler;
            return;
         end if;
      end loop;

      raise Out_Of_Memory;

   exception
      when The_Error : Others =>
         raise;
   end Register;

   ---=====================================================================---
   ---===                        M E T H O D S                          ===---
   ---=====================================================================---

end ASCL.RT.Resource;
