--|                              -*- Mode: Ada -*-
--| Filename        : $Source: /home/ACDE/ACDE/ASCL/RCS/ascl-rt-resource.ads,v $
--| Description     : Watched Resources
--| Author          : Michael Erdmann
--| Created On      : 25.3.1999
--| Last Modified By: $Author: ACDE $
--| Last Modified On: $Date: 2000/01/23 09:48:34 $
--| Status          : $State: Released $
--|
--| Functional Description
--| ======================
--| This component provides so called watched resources. This means
--| if any instance of an derived type if initialized it will be 
--| Monitored that the resource is not kept longer then the
--| specified time.
--|
--| Each implmentation of this class has to register an error 
--| handler which is called if an resource error has been 
--| found. The error handler is called with the resource id
--| as an argument. It it a job of the error handler to take
--| the apropiate actions.
--|
--|
--| Component Data
--| ==============
--| Resource_ID   - Each controled resource is identified by a resource
--|                 identifier. The resource identifier is allocated at
--|                 the instanciation time and is only valid during
--|                 the life time of the object.
--|
--|
--| Error Handling
--| ==============
--|
--| No exceptions are generated by this component.
--|
--| Extension
--| =========
--|
--| Deallocate
--|
--|    Is called upon finalization of the object. This may be used to 
--|    clean up some local object tables etc...
--|
--| Restrictions
--| ============
--|
--| References
--| ==========
--|
--| #$-DOCUMENT: Example Source-$#
--|
--|
--| History
--| =======
--| $Log: ascl-rt-resource.ads,v $
--| Revision 1.2  2000/01/23 09:48:34  ACDE
--| Release Version base_1_1
--|
--| Revision 1.1  2000/01/23 09:46:34  ACDE
--| Release Version 1_0
--|
--| Revision 1.11  1999/11/27 16:56:18  ACDE
--| Release Version 1P010
--|
--| Revision 1.1  1999/11/27 09:33:55  ACDE
--| sg
--|
--| Revision 1.3  1999/08/01 19:14:22  erdmann
--| Added debuging support (ASCL.Debugging_SUpport)
--|
--| Revision 1.2  1999/07/26 17:42:37  erdmann
--| Tesing completed, see test-resource, main
--|
--| Revision 1.1  1999/07/25 10:10:35  erdmann
--| No comments
--|
--|
with Ada.Finalization;  use Ada.Finalization;
with Ada.Tags;          use Ada.Tags;

with ASCL.Debugging_Support;   

use ASCL;

package ASCL.RT.Resource is
   
   ---=====================================================================---
   ---===             C O M P O N E N T   I N T E R F A C E             ===---
   ---=====================================================================---
   type Object( time : Natural ) is abstract tagged limited private;
   type Handle is access all Object'Class;

   procedure Initialize( this  : in out Object'Class;
                         debug : in Debugging_Support.Handle := null );

   procedure Finalize( this : in out Object'Class );

   ---=====================================================================---
   ---===                  A T T R I B U T E S                       ===---
   ---=====================================================================---

   type Resource_ID is new Natural range 0..2000;
   Null_Resource_ID : constant Resource_ID := 0;

   function Id( this : in Object ) return Resource_ID;

   type Resource_Handler is access procedure( id : Resource_ID );
   procedure Register( name : in Tag; Handler : in Resource_Handler );

   ---=====================================================================---
   ---===                     M E T H O D S                             ===---
   ---=====================================================================---

   ---=====================================================================---
   ---===                   E X T E N S I O N                           ===---
   ---=====================================================================---
   procedure Deallocate( this : in Object )                    is abstract;
 
   ---=====================================================================---
private
   
   type Watch_Dog( super : access Object'Class ) is 
             new Limited_Controlled with null record;

   procedure Initialize( this : in out Watch_Dog );
   procedure Finalize( this : in out Watch_Dog );

   type Object( time : Natural ) is  abstract tagged limited record
         id        : Resource_ID := 0;
         life_time : Natural     := time;
         data      : Watch_Dog( Object'Access ) ;           
      end record;

end ASCL.RT.Resource;
